﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Reflection;

namespace DragDataGridView
{
    public delegate void OnDragDropCompletedHandler(object sender, DragDataGridDragCompletedEventArgs e);

    public partial class DragDataGrid : UserControl
    {
        private bool IsDragging { get; set; }
        private bool IsCopy { get; set; }
        private bool IsEditing { get; set; }
        private List<Animal> _dataSource;

        public List<Animal> DataSource
        {
            get { return _dataSource; }
            set
            {
                _dataSource = value;
                this.dgWPFAnimalInfo.ItemsSource = value;
            }
        }

        public event OnDragDropCompletedHandler OnDragDropCompleted;

        #region DraggedItem

        public static readonly DependencyProperty DraggedItemProperty =
            DependencyProperty.Register("DraggedItem", typeof(Animal), typeof(DragDataGrid));

        public Animal DraggedItem
        {
            get { return (Animal)GetValue(DraggedItemProperty); }
            set { SetValue(DraggedItemProperty, value); }
        }

        #endregion

        public DragDataGrid()
        {
            InitializeComponent();
        }

        #region edit mode monitoring

        private void OnBeginEdit(object sender, DataGridBeginningEditEventArgs e)
        {
            IsEditing = true;
            if (IsDragging)
                ResetDragDrop();
        }

        private void OnEndEdit(object sender, DataGridCellEditEndingEventArgs e)
        {
            IsEditing = false;
        }

        #endregion

        #region Drag and Drop Rows

        private void OnMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            if (IsEditing)
                return;

            var row = UIHelpers.TryFindFromPoint<DataGridRow>((UIElement)sender, e.GetPosition(this.dgWPFAnimalInfo));
            if (row == null || row.IsEditing)
                return;
            IsDragging = true;
            DraggedItem = (Animal)row.Item;

            Image imgIcon = popup1.FindName("imgIcon") as Image;
            popup1.DataContext = row.Item as Animal;
            if (Keyboard.IsKeyDown(Key.LeftCtrl) || Keyboard.IsKeyDown(Key.RightCtrl))
            {
                IsCopy = true;
                imgIcon.Source = new BitmapImage(new Uri(@"Images/ImgMove.png", UriKind.Relative));
            }
            else
            {
                IsCopy = false;
                imgIcon.Source = new BitmapImage(new Uri(@"Images/ImgCopy.png", UriKind.Relative));
            }
        }

        private void OnMouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            if (!IsDragging || IsEditing)
                return;

            Animal targetItem = (Animal)this.dgWPFAnimalInfo.SelectedItem;

            if (targetItem == null || !ReferenceEquals(DraggedItem, targetItem))
            {
                var targetIndex = DataSource.IndexOf(targetItem);

                Animal tmpObj = new Animal();
                CopyModel<Animal>(DraggedItem, tmpObj);

                if (!IsCopy)
                    DataSource.Remove(DraggedItem);
                DataSource.Insert(targetIndex, tmpObj);
                this.dgWPFAnimalInfo.ItemsSource = null;
                this.dgWPFAnimalInfo.ItemsSource = DataSource;
                this.dgWPFAnimalInfo.SelectedItem = tmpObj;

                if (this.OnDragDropCompleted != null)
                {
                    DragDataGridDragCompletedEventArgs eventArgs = new DragDataGridDragCompletedEventArgs();
                    eventArgs.DragEffect = IsCopy ? DragEffects.Copy : DragEffects.Move;
                    eventArgs.DragItem = tmpObj;
                    this.OnDragDropCompleted(this, eventArgs);
                }
            }
            ResetDragDrop();
        }

        //Deep Copy
        private void CopyModel<T>(T oriEntity, T newEntity)
        {
            if (oriEntity == null || newEntity == null)
                return;
            Type type = typeof(T);
            PropertyInfo[] proInfoArray = type.GetProperties();
            foreach (PropertyInfo proInfo in proInfoArray)
            {
                if (proInfo.CanWrite)
                {
                    object oriObj = type.GetProperty(proInfo.Name).GetValue(oriEntity, null);
                    type.GetProperty(proInfo.Name).SetValue(newEntity, oriObj, null);
                }
            }
        }

        private void ResetDragDrop()
        {
            IsDragging = false;
            popup1.IsOpen = false;
            this.dgWPFAnimalInfo.IsReadOnly = false;
        }

        private void OnMouseMove(object sender, MouseEventArgs e)
        {
            if (!IsDragging || e.LeftButton != MouseButtonState.Pressed) return;

            if (!popup1.IsOpen)
            {
                this.dgWPFAnimalInfo.IsReadOnly = true;
                popup1.IsOpen = true;
            }

            Size popupSize = new Size(popup1.ActualWidth, popup1.ActualHeight);
            popup1.PlacementRectangle = new Rect(e.GetPosition(this), popupSize);
            Point position = e.GetPosition(this.dgWPFAnimalInfo);
            var row = UIHelpers.TryFindFromPoint<DataGridRow>(this.dgWPFAnimalInfo, position);
            if (row != null)
                this.dgWPFAnimalInfo.SelectedItem = row.Item;
        }

        #endregion

        private void dgWPFAnimalInfo_Drop(object sender, DragEventArgs e)
        {
            string strRow = (string)e.Data.GetData(DataFormats.StringFormat);
            Animal aimAnimal = (Animal)DataSerializer.DeserializeObject(strRow);

            Animal targetItem = (Animal)this.dgWPFAnimalInfo.SelectedItem;
            var targetIndex = DataSource.IndexOf(targetItem);

            DataSource.Insert(targetIndex, aimAnimal);
            this.dgWPFAnimalInfo.ItemsSource = null;
            this.dgWPFAnimalInfo.ItemsSource = DataSource;
            this.dgWPFAnimalInfo.SelectedItem = aimAnimal;

            if (this.OnDragDropCompleted != null)
            {
                DragDataGridDragCompletedEventArgs eventArgs = new DragDataGridDragCompletedEventArgs();
                eventArgs.DragEffect = IsCopy ? DragEffects.Copy : DragEffects.Move;
                eventArgs.DragItem = aimAnimal;
                this.OnDragDropCompleted(this, eventArgs);
            }
        }

        private void dgWPFAnimalInfo_DragOver(object sender, DragEventArgs e)
        {
            var row = UIHelpers.TryFindFromPoint<DataGridRow>((UIElement)sender, e.GetPosition(this.dgWPFAnimalInfo));
            if (row != null)
            {
                this.dgWPFAnimalInfo.SelectedItem = row.Item;
            }
            string strRow = (string)e.Data.GetData(DataFormats.StringFormat);
            Animal aimAnimal = (Animal)DataSerializer.DeserializeObject(strRow);
        }
    }

    public class DragDataGridDragCompletedEventArgs
    {
        public object DragItem { get; set; }
        public DragEffects DragEffect { get; set; }
    }

    public enum DragEffects
    {
        Move,
        Copy
    }
}
